
/******************************************************************************
* MODULE     : real.cpp
* DESCRIPTION: Effective real numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "real.hpp"
namespace mml {

/******************************************************************************
* Improve the precision of the next approximation as a function of
* the complexity of the approximation algorithm (linear or similar to
* the complexity of multiplication).
******************************************************************************/

mp_prec_t
real_rep::linear_algo (const dyadic& err) {
  mp_exp_t bound_expo, cur_prec, algo_prec, req_prec, best_step, err_expo;
  if (is_fuzzy (best)) {
    bound_expo= expo (get_bound ());
    cur_prec  = 0;
  }
  else {
    bound_expo= expo (bound (this));
    best_step = step (best);
    cur_prec  = max ((mp_exp_t) 0, (mp_exp_t) bound_expo - best_step);
  }
  if (is_infinity (err))
    req_prec  = DEFAULT_LOW;
  else {
    err_expo  = expo (err);
    req_prec  = bound_expo + DEFAULT_LOW - err_expo;
  }
  algo_prec = (cur_prec << 1) + DEFAULT_LOW;
  return mpfr_prec_round_up (max (algo_prec, req_prec));
}

mp_prec_t
real_rep::mult_algo (const dyadic& err) {
  mp_exp_t bound_expo, cur_prec, algo_prec, req_prec, best_step, err_expo;
  if (is_fuzzy (best)) {
    bound_expo= expo (get_bound ());
    cur_prec= 0;
  }
  else {
    bound_expo= expo (bound (this));
    best_step = step (best);
    cur_prec  = max ((mp_exp_t) 0, (mp_exp_t) bound_expo - best_step);
  }
  if (is_infinity (err))
    req_prec  = DEFAULT_LOW;
  else {
    err_expo  = expo (err);
    req_prec  = bound_expo + DEFAULT_LOW - err_expo;
  }
  algo_prec = cur_prec + (cur_prec >> 1);
  if (cur_prec >= (BITS_PER_LIMB << 8)) algo_prec= cur_prec << 1;
  return mpfr_prec_round_up (max (algo_prec, req_prec));
}

/******************************************************************************
* Normalization of non-zero real numbers
******************************************************************************/

bool
real_rep::normalize (mp_prec_t max_cancel, const dyadic& rel_err) {
  (void) approx (lp_infinity);
  while (true) {
    if (radius (best) < rel_err * abs (center (best))) return false;
    if (is_epsilon (best)) return true;
    if (abs (center (best)) < (get_bound () >> max_cancel)) return true;
    (void) approx (radius (best) >> BITS_PER_LIMB);
  }
}

int
sign (const real& x, int max_cancel= MAX_CANCEL_TEST) {
  if (x->normalize (max_cancel))
    /*cerr << "Warning: heuristic determination of sign of " << x << "\n"*/;
  if (x->best > 0) return 1;
  if (x->best < 0) return -1;
  return 0;
}

}
